from typing import Optional, Union, List, Tuple

from pyjsg.parser.jsgParser import *
from pyjsg.parser.jsgParserVisitor import jsgParserVisitor
from pyjsg.parser_impl.jsg_doc_context import JSGDocContext, PythonGeneratorElement
from pyjsg.parser_impl.jsg_ebnf_parser import JSGEbnf
from pyjsg.parser_impl.jsg_lexerruleblock_parser import JSGLexerRuleBlock
from pyjsg.parser_impl.jsg_pairdef_parser import JSGPairDef
from pyjsg.parser_impl.jsg_valuetype_parser import JSGValueType
from pyjsg.parser_impl.parser_utils import t, flatten
from .parser_utils import as_token, flatten_unique, is_valid_python

_class_template = """


class {name}(jsg.JSGObject):
    _reference_types = [{reference_types}]
    _members = {{{members}}}
    _strict = {strict}

{init_fctn}"""

_map_template = """


class {name}(jsg.JSGObjectMap):{name_filter}{value_type}

    def __init__(self,
                 **_kwargs):
        super().__init__(_CONTEXT, **_kwargs)"""


_init_template = """    def __init__(self{signatures},
                 **_kwargs: typing.Dict[str, object]):
        super().__init__(_CONTEXT, **_kwargs){initializers}
"""
indent0 = ",\n                "
indent1n = "\n    "
indent1 = indent0 + " "
indent2 = "\n        "
indent3 = indent2 + "    "


class JSGObjectExpr(jsgParserVisitor, PythonGeneratorElement):
    """ objectExpr: OBRACE membersDef? CBRACE | OBRACE ID MAPSTO valueType CBRACE """
    def __init__(self, context: JSGDocContext,
                 ctx: Optional[Union[jsgParser.ObjectExprContext, jsgParser.MembersDefContext]] = None,
                 name: Optional[str] = None):
        self._context = context
        self._name = name
        self._strict = True

        # _members, _choices and _map_name_type are mutually exclusive
        self._members: List[JSGPairDef] = []
        self._choices: List[str] = []

        # _map is for a map style definition
        self._map_name_type: Optional[Union[str, JSGLexerRuleBlock]] = None               # Name of a lexer rule block
        self._map_valuetype: Optional[JSGValueType] = None
        self._map_ebnf: Optional[JSGEbnf] = None
        self.text = ""

        if ctx:
            self.text = ctx.getText()
            self.visit(ctx)

    def __str__(self):
        return "objectExpr: {}" \
            .format("object map" if self._map_name_type else "object choices" if self._choices else "simple object")

    def map_as_python(self, name: str) -> str:
        if self._map_name_type is not None:
            if isinstance(self._map_name_type, str):
                filtr = self._map_name_type
            else:
                filtr = self._map_name_type.signature_type()
            name_filter = f"{indent1n}_name_filter = {filtr}"
        else:
            name_filter = ""
        if self._map_valuetype is not None:
            typ = self._map_valuetype.python_type()
        else:
            typ = "jsg.AnyType"
        value_type = f"{indent1n}_value_type = {typ}"
        return _map_template.format(**locals())

    def as_python(self, name: str) -> str:
        """ Return the python representation of the class represented by this object """
        if self._map_valuetype:
            return self.map_as_python(name)
        else:
            return self.obj_as_python(name)

    def obj_as_python(self, name: str) -> str:
        reference_types = self._gen_reference_types()
        members = self._gen_members_list()
        strict = self._strict
        init_fctn = self._gen_init_fctn()
        return _class_template.format(**locals())

    def _gen_reference_types(self) -> str:
        from pyjsg.jsglib import AnyType
        if self._map_valuetype:
            return self._map_valuetype if self._map_valuetype is not AnyType else ''
        elif self._choices:
            return ', '.join(r for r in self._choices)
        else:
            return ', '.join(pair.python_base_type() for pair in self._members if pair.is_reference_type())

    def _gen_members_list(self) -> str:
        if self._map_valuetype:
            return ''
        else:
            return indent0.join(f"'{mk}': {mt}" for mk, mt in self.members_entries())

    def _gen_init_fctn(self) -> str:
        pair_signatures = self.signatures()
        pair_initializers = self.initializers()
        if self._map_valuetype:
            return ''
        elif self._choices:
            signatures = indent1 + indent1.join(pair_signatures) if pair_signatures else ""
            initializers = indent2 + indent2.join(pair_initializers)
            return _init_template.format(**locals())
        else:
            signatures = indent1 + indent1.join(self.signatures()) if pair_signatures else ""
            initializers = indent2 + indent2.join(flatten(pair_initializers)) if pair_initializers else ""
            return _init_template.format(**locals())

    def python_type(self) -> str:
        return self._name

    def signature_type(self) -> str:
        return self._name

    def reference_type(self):
        return self._name

    def mt_value(self) -> str:
        return 'None'

    def signatures(self, _: bool=False) -> List[str]:
        if self._map_valuetype:
            return []
        elif self._choices:
            pair_signatures = ', '.join(self._choices)
            if len(self._choices) > 1:
                pair_signatures = f'opts_: typing.Union[{pair_signatures}] = None'
            else:
                pair_signatures = f'{self._choices[0]}: {pair_signatures} = None'
            return [pair_signatures]
        else:
            return flatten([pairdef.signatures() for pairdef in self._members])

    def initializers(self, prefix: Optional[str] = None) -> List[str]:

        if self._map_valuetype:
            return []
        elif self._choices:
            if not prefix:
                prefix = 'opts_'
            # Overall test
            rval = [f'if {prefix} is not None:']

            # First choice option
            choice = self._choices[0]
            rval.append(f'{t(1)}if isinstance({prefix}, {choice}):')
            choice_ref = self._context.reference(choice)
            inits = choice_ref.initializers(prefix)
            for initializer in inits:
                rval.append(t(2) + initializer)
            if not inits:
                rval.append(t(2) + 'pass')

            # Second through nth choice options
            for choice in self._choices[1:]:
                if is_valid_python(choice):
                    rval.append(f'{t(1)}elif isinstance({prefix}, {choice}):')
                    choice_ref = self._context.reference(choice)
                    choice_inits = []
                    for initializer in choice_ref.initializers(prefix):
                        choice_inits.append(t(2) + initializer)
                    if not choice_inits:
                        choice_inits.append(t(2) + 'pass')
                    rval += choice_inits
                else:
                    raise NotImplementedError("Can this happen?")

            # Error checker
            rval.append(t() + 'else:')
            rval.append(t(2) + f'raise ValueError(f"Unrecognized value type: {{{prefix}}}")')
            return rval
        else:
            return flatten([pairdef.initializers(prefix) for pairdef in self._members])

    def members_entries(self, all_are_optional: bool=False) -> List[Tuple[str, str]]:
        """ Return an ordered list of elements for the _members section

        :param all_are_optional: True means we're in a choice situation so everything is optional
        :return:
        """
        rval = []
        if self._members:
            for member in self._members:
                rval += member.members_entries(all_are_optional)
        elif self._choices:
            for choice in self._choices:
                rval += self._context.reference(choice).members_entries(True)
        else:
            return []
        return rval

    def dependency_list(self) -> List[str]:
        if self._members:
            return flatten_unique(member.dependency_list() for member in self._members)
        elif self._map_valuetype:
            return self._map_valuetype.dependency_list()
        else:
            return flatten_unique(self._context.dependency_list(c) + [c] for c in self._choices)

    # ***************
    #   Visitors
    # ***************
    def visitObjectExpr(self, ctx: jsgParser.ObjectExprContext):
        """ objectExpr: OBRACE membersDef? CBRACE 
                        OBRACE (LEXER_ID_REF | ANY)? MAPSTO valueType ebnfSuffix? CBRACE
        """
        if not self._name:
            self._name = self._context.anon_id()
        if ctx.membersDef():
            self.visitChildren(ctx)
        elif ctx.MAPSTO():
            if ctx.LEXER_ID_REF():
                self._map_name_type = as_token(ctx)
            # Any and absent mean the same thing
            self._map_valuetype = JSGValueType(self._context, ctx.valueType())
            if ctx.ebnfSuffix():
                self._map_ebnf = JSGEbnf(self._context, ctx.ebnfSuffix())

    def visitMembersDef(self, ctx: jsgParser.MembersDefContext):
        """ membersDef: COMMA | member+ (BAR altMemberDef)* (BAR lastComma)? ;
            altMemberDef: member* ;
            member: pairDef COMMA?
            lastComma: COMMA ;
        """
        if not self._name:
            self._name = self._context.anon_id()
        if ctx.COMMA():                             # lone comma - wild card
            self._strict = False
        if not ctx.BAR():                           # member+
            self.visitChildren(ctx)
        else:
            entry = 1
            self._add_choice(entry, ctx.member())       # add first brance (member+)
            for alt in ctx.altMemberDef():
                entry += 1
                self._add_choice(entry, alt.member())
            if ctx.lastComma():
                entry += 1
                self._add_choice(entry, [])

    def visitMember(self, ctx: jsgParser.MemberContext):
        """ member: pairDef COMMA? """
        self._strict = ctx.COMMA() is None
        self.visitChildren(ctx)

    def visitPairDef(self, ctx: jsgParser.PairDefContext):
        self._members.append(JSGPairDef(self._context, ctx))

    def _add_choice(self, branch: int, ctx: List[jsgParser.MemberContext]):
        choice_name = "{}_{}_".format(self._name, branch)
        choice_obj = JSGObjectExpr(self._context, name=choice_name)
        for member in ctx:
            choice_obj.visit(member)
        self._choices.append(choice_name)
        self._context.grammarelts[choice_name] = choice_obj



import importlib.abc
import importlib.util
import os
from argparse import ArgumentParser
from typing import Optional, Union

import sys
from antlr4 import CommonTokenStream
from antlr4 import FileStream, InputStream
from antlr4.error.ErrorListener import ErrorListener
from pyjsg.parser.jsgParser import jsgParser

from pyjsg.parser.jsgLexer import jsgLexer
from pyjsg.parser_impl.jsg_doc_parser import JSGDocParser


class ParseErrorListener(ErrorListener):

    def __init__(self):
        self.n_errors = 0

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self.n_errors += 1

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        print("*** Parsing ambiguity error", file=sys.stderr)
        self.n_errors += 1

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        print("*** Parsing full context error", file=sys.stderr)
        self.n_errors += 1

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        print("*** Context sensitivity error", file=sys.stderr)
        self.n_errors += 1


def do_parse(infilename: str, outfilename: str, verbose: bool) -> bool:
    """
    Parse the jsg in infilename and save the results in outfilename
    :param infilename: file containing jsg
    :param outfilename: target python file
    :param verbose: verbose output flag
    :return: true if success
    """
    python = parse(FileStream(infilename, encoding="utf-8"), infilename)
    if python is not None:
        with open(outfilename, 'w') as outfile:
            outfile.write(python)
        if verbose:
            print("Output written to {}".format(outfilename))
        return True
    return False


def parse(input_: Union[str, FileStream], source: str) -> Optional[str]:
    """Parse the text in infile and save the results in outfile

    :param input_: string or stream to parse
    :param source: source name for python file header
    :return: python text if successful
    """

    # Step 1: Tokenize the input stream
    error_listener = ParseErrorListener()
    if not isinstance(input_, FileStream):
        input_ = InputStream(input_)
    lexer = jsgLexer(input_)
    lexer.addErrorListener(error_listener)
    tokens = CommonTokenStream(lexer)
    tokens.fill()
    if error_listener.n_errors:
        return None

    # Step 2: Generate the parse tree
    parser = jsgParser(tokens)
    parser.addErrorListener(error_listener)
    parse_tree = parser.doc()
    if error_listener.n_errors:
        return None

    # Step 3: Transform the results the results
    parser = JSGDocParser()
    parser.visit(parse_tree)

    if parser.undefined_tokens():
        for tkn in parser.undefined_tokens():
            print("Undefined token: " + tkn)
        return None

    return parser.as_python(source)


def genargs() -> ArgumentParser:
    """
    Create a command line parser
    :return: parser
    """
    parser = ArgumentParser()
    parser.add_argument("infile", help="Input JSG specification")
    parser.add_argument("-o", "--outfile", help="Output python file (Default: {infile}.py)")
    parser.add_argument("-e", "--evaluate", help="Evaluate resulting python file as a test", action="store_true")
    parser.add_argument("-v", "--verbose", help="Verbose output", action="store_true")
    return parser


def evaluate(module_name: str, fname: str, verbose: bool):
    """
    Load fname as a module.  Will raise an exception if there is an error
    :param module_name: resulting name of module
    :param fname: name to load 
    """
    if verbose:
        print("Testing {}".format(fname))
    spec = importlib.util.spec_from_file_location(module_name, fname)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)


def generate(argv) -> bool:
    opts = genargs().parse_args(argv)
    file_base = str(os.path.basename(opts.infile.rsplit('.', 1)[0]))
    if not opts.outfile:
        opts.outfile = os.path.join(os.path.dirname(opts.infile), file_base + ".py")
    if do_parse(opts.infile, opts.outfile, opts.verbose):
        if opts.evaluate:
            evaluate("generate_python_namespace", opts.outfile, opts.verbose)               # Don't pollute namespace
        return True
    else:
        print("Conversion failed")
        return False



from dataclasses import dataclass
from typing import Optional, Tuple, Dict, List

from pyjsg.parser.jsgParser import *
from pyjsg.parser.jsgParserVisitor import jsgParserVisitor
from pyjsg.parser_impl.jsg_doc_context import JSGDocContext, PythonGeneratorElement


class JSGBuiltinValueType(jsgParserVisitor, PythonGeneratorElement):
    @dataclass
    class TypeInfo:
        sig_type: str
        ref_type: str
        py_type: type
        mt_value: str

    parserTypeToImplClass: Dict[str, TypeInfo] = \
        {"@string": TypeInfo("jsg.String", "jsg.String", str, "None"),
         "@object": TypeInfo("jsg.ObjectFactory('{name}', _CONTEXT, jsg.Object)", "jsg.JSGObject", object, "None"),
         "@int": TypeInfo("jsg.Integer", "jsg.Integer", int, "None"),
         "@number": TypeInfo("jsg.Number", "jsg.Number", float, "None"),
         "@null": TypeInfo("jsg.JSGNull", "jsg.JSGNull", type(None), "jsg.Empty"),
         "@array": TypeInfo("jsg.ArrayFactory('{name}', _CONTEXT, jsg.AnyType, 0, None)", "jsg.JSGArray", list, "None"),
         "@bool": TypeInfo("jsg.Boolean", "jsg.Boolean", bool, "None"),
         ".": TypeInfo("jsg.AnyTypeFactory('{name}', _CONTEXT)", "jsg.AnyType", object, "jsg.Empty")}

    def __init__(self, context: JSGDocContext, ctx: Optional[jsgParser.BuiltinValueTypeContext] = None):
        self._context = context
        self._value_type_text: Optional[str] = None
        self._typeinfo: JSGBuiltinValueType.TypeInfo = None
        self.text = ""
        if ctx:
            self.text = ctx.getText()
            self.visit(ctx)

    def __str__(self):
        return f"builtinValueType: {self._value_type_text if self._value_type_text != '.' else 'jsg.AnyType'}"

    def python_type(self) -> str:
        return "type(None)" if self._typeinfo.py_type is type(None) else self._typeinfo.py_type.__name__

    def signature_type(self) -> str:
        return self._typeinfo.sig_type

    def reference_type(self) -> str:
        return self._typeinfo.sig_type

    def mt_value(self) -> str:
        return self._typeinfo.mt_value

    def members_entries(self, all_are_optional: Optional[bool] = False) -> List[Tuple[str, str]]:
        return []

    def dependency_list(self) -> List[str]:
        return []

    # ***************
    #   Visitors
    # ***************
    def visitBuiltinValueType(self, ctx: jsgParser.BuiltinValueTypeContext):
        """ valueTypeExpr: JSON_STRING | JSON_NUMBER | JSON_INT | JSON_BOOL | JSON_NULL | JSON_ARRAY | JSON_OBJECT """
        self._value_type_text = ctx.getText()
        self._typeinfo = self.parserTypeToImplClass[self._value_type_text]



# Generated from jsgParser.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .jsgParser import jsgParser
else:
    from jsgParser import jsgParser

# This class defines a complete generic visitor for a parse tree produced by jsgParser.

class jsgParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by jsgParser#doc.
    def visitDoc(self, ctx:jsgParser.DocContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#typeDirective.
    def visitTypeDirective(self, ctx:jsgParser.TypeDirectiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#typeExceptions.
    def visitTypeExceptions(self, ctx:jsgParser.TypeExceptionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#ignoreDirective.
    def visitIgnoreDirective(self, ctx:jsgParser.IgnoreDirectiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#grammarElt.
    def visitGrammarElt(self, ctx:jsgParser.GrammarEltContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#objectDef.
    def visitObjectDef(self, ctx:jsgParser.ObjectDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#objectExpr.
    def visitObjectExpr(self, ctx:jsgParser.ObjectExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#membersDef.
    def visitMembersDef(self, ctx:jsgParser.MembersDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#altMemberDef.
    def visitAltMemberDef(self, ctx:jsgParser.AltMemberDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#member.
    def visitMember(self, ctx:jsgParser.MemberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lastComma.
    def visitLastComma(self, ctx:jsgParser.LastCommaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#pairDef.
    def visitPairDef(self, ctx:jsgParser.PairDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#name.
    def visitName(self, ctx:jsgParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#arrayDef.
    def visitArrayDef(self, ctx:jsgParser.ArrayDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#arrayExpr.
    def visitArrayExpr(self, ctx:jsgParser.ArrayExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#objectMacro.
    def visitObjectMacro(self, ctx:jsgParser.ObjectMacroContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#valueTypeMacro.
    def visitValueTypeMacro(self, ctx:jsgParser.ValueTypeMacroContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#builtinValueType.
    def visitBuiltinValueType(self, ctx:jsgParser.BuiltinValueTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#valueType.
    def visitValueType(self, ctx:jsgParser.ValueTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#nonRefValueType.
    def visitNonRefValueType(self, ctx:jsgParser.NonRefValueTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#typeAlternatives.
    def visitTypeAlternatives(self, ctx:jsgParser.TypeAlternativesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#idref.
    def visitIdref(self, ctx:jsgParser.IdrefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#ebnfSuffix.
    def visitEbnfSuffix(self, ctx:jsgParser.EbnfSuffixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerRules.
    def visitLexerRules(self, ctx:jsgParser.LexerRulesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerRuleSpec.
    def visitLexerRuleSpec(self, ctx:jsgParser.LexerRuleSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerRuleBlock.
    def visitLexerRuleBlock(self, ctx:jsgParser.LexerRuleBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerAltList.
    def visitLexerAltList(self, ctx:jsgParser.LexerAltListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerAlt.
    def visitLexerAlt(self, ctx:jsgParser.LexerAltContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerElements.
    def visitLexerElements(self, ctx:jsgParser.LexerElementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerElement.
    def visitLexerElement(self, ctx:jsgParser.LexerElementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerBlock.
    def visitLexerBlock(self, ctx:jsgParser.LexerBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerAtom.
    def visitLexerAtom(self, ctx:jsgParser.LexerAtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by jsgParser#lexerTerminal.
    def visitLexerTerminal(self, ctx:jsgParser.LexerTerminalContext):
        return self.visitChildren(ctx)



del jsgParser


from typing import Optional, List, Union, cast, Tuple

from pyjsg.parser.jsgParser import *
from pyjsg.parser.jsgParserVisitor import jsgParserVisitor
from pyjsg.parser_impl.jsg_arrayexpr_parser import JSGArrayExpr
from pyjsg.parser_impl.jsg_builtinvaluetype_parser import JSGBuiltinValueType
from pyjsg.parser_impl.jsg_doc_context import JSGDocContext, PythonGeneratorElement
from pyjsg.parser_impl.parser_utils import as_token, flatten_unique


class JSGValueType(jsgParserVisitor, PythonGeneratorElement):
    """ Parser for valueType production """

    def __init__(self, context: JSGDocContext,
                 ctx: Optional[Union[jsgParser.ValueTypeContext, jsgParser.ValueTypeMacroContext,
                                     jsgParser.NonRefValueTypeContext]] = None):
        self._context = context

        # With the exception of lexeridref and alttypelist, all options below are mutually exclusive
        self._typeid: Optional[str] = None              # The name of a referenced (possibly anonymous) type
        self._arrayDef: Optional[JSGArrayExpr] = None   # Type is an array
        self._lexeridref: Optional[str] = None          # The name of a (possibly anonymous) string match pattern
        self._builtintype: Optional[JSGBuiltinValueType] = None  # Type is a builtin type
        self._alttypelist: List[JSGValueType] = []      # Two or more alternative types
        self.text = ""

        if ctx:
            self.text = ctx.getText()
            self.visit(ctx)

    def __str__(self):
        if self._typeid:
            if self._context.is_anon(self._typeid):
                typ = f"(anonymous: {self._typeid}): {self._context.reference(self._typeid)}"
            else:
                typ = f"ID: {self._typeid}"
        elif self._builtintype:
            typ = str(self._builtintype)
        elif self._alttypelist:
            lid_str = f"({self._lexerid_str()}) | " if self._lexeridref else ""
            alts_str = ' | '.join(ta.signature_type() for ta in self._alttypelist)
            typ = f"({lid_str}{alts_str})"
        elif self._lexeridref:
            typ = self._lexerid_str()
        elif self._arrayDef:
            typ = str(self._arrayDef)
        else:
            raise NotImplementedError("Unknown Value Type")
        return "valueType: {}".format(typ)

    def _lexerid_str(self) -> str:
        if not self._context.is_anon(self._lexeridref):
            return "LEXER_ID_REF: {}".format(self._lexeridref)
        else:
            return "STRING: {}".format(self._context.grammarelts[self._lexeridref])

    def python_type(self) -> str:
        """ Return the official python type for the value. As an example, an '@int' aps to 'int' or a match pattern
        maps to 'str'
        """
        types = []
        if self._lexeridref:
            types.append('str')
        if self._typeid:
            types.append(self._context.python_type(self._typeid))
        if self._builtintype:
            types.append(self._builtintype.python_type())
        if self._alttypelist:
            types += [e.python_type() for e in self._alttypelist]
        if self._arrayDef:
            types.append(self._arrayDef.python_type())
        return "jsg.AnyType" if len(types) == 0 else \
            types[0] if len(types) == 1 else "typing.Union[{}]".format(', '.join(types))

    def signature_type(self) -> str:
        """ Return the signature type for the value. As an example, an '@int' maps to 'Integer' or a match pattern
        maps to the name of the JSGString class """
        types = []
        if self._lexeridref:
            types.append(self._lexeridref)
        if self._typeid:
            types.append(self._context.signature_type(self._typeid))
        if self._builtintype:
            types.append(self._builtintype.signature_type())
        if self._alttypelist:
            types += [e.signature_type() for e in self._alttypelist]
        if self._arrayDef:
            types.append(self._arrayDef.signature_type())
        return "jsg.AnyType" if len(types) == 0 else \
            types[0] if len(types) == 1 else "typing.Union[{}]".format(', '.join(types))

    def reference_type(self) -> str:
        types = []
        if self._lexeridref:
            types.append(self._lexeridref)
        if self._typeid:
            types.append(self._context.reference_type(self._typeid))
        if self._builtintype:
            types.append(self._builtintype.reference_type())
        if self._alttypelist:
            types += [e.reference_type() for e in self._alttypelist]
        if self._arrayDef:
            types.append(self._arrayDef.reference_type())
        return "jsg.AnyType" if len(types) == 0 else \
            types[0] if len(types) == 1 else "typing.Union[{}]".format(', '.join(types))


    def mt_value(self) -> str:
        return self._builtintype.mt_value() if self._builtintype else "None"


    def dependency_list(self) -> List[str]:
        rval = []
        if self._typeid:
            rval.append(self._typeid)
            rval += self._context.dependency_list(self._typeid)
        if self._lexeridref:
            rval.append(self._lexeridref)
        if self._arrayDef:
            rval += flatten_unique((self._arrayDef.dependency_list()))
        if self._alttypelist:
            rval += flatten_unique([e.dependency_list() for e in self._alttypelist])
        return rval

    def members_entries(self, all_are_optional: Optional[bool] = False) -> List[Tuple[str, str]]:
        return []

    # ***************
    #   Visitors
    # ***************
    def visitValueType(self, ctx: jsgParser.ValueTypeContext):
        """ valueType: idref | nonRefValueType """
        if ctx.idref():
            self._typeid = as_token(ctx)
        else:
            self.visitChildren(ctx)

    def visitNonRefValueType(self, ctx: jsgParser.NonRefValueTypeContext):
        """ nonRefValueType: LEXER_ID_REF | STRING | builtinValueType | objectExpr | arrayExpr  
                             | OPREN typeAlternatives CPREN | ANY """
        if ctx.LEXER_ID_REF():                  # Reference to a lexer token
            self._lexeridref = as_token(ctx)
        elif ctx.STRING():                      # Anonymous lexer token
            from pyjsg.parser_impl.jsg_lexerruleblock_parser import JSGLexerRuleBlock
            lrb = JSGLexerRuleBlock(self._context)
            lrb.add_string(ctx.getText()[1:-1], False)
            self._lexeridref = self._context.anon_id()
            self._context.grammarelts[self._lexeridref] = lrb
        else:
            self.visitChildren(ctx)

    def visitBuiltinValueType(self, ctx: jsgParser.BuiltinValueTypeContext):
        from pyjsg.parser_impl.jsg_builtinvaluetype_parser import JSGBuiltinValueType
        self._builtintype = JSGBuiltinValueType(self._context, ctx)

    def visitObjectExpr(self, ctx: jsgParser.ObjectExprContext):
        from pyjsg.parser_impl.jsg_objectexpr_parser import JSGObjectExpr
        oe = JSGObjectExpr(self._context, ctx)
        self._typeid = oe.python_type()
        self._context.grammarelts[self._typeid] = oe

    def visitArrayExpr(self, ctx: jsgParser.ArrayExprContext):
        from pyjsg.parser_impl.jsg_arrayexpr_parser import JSGArrayExpr
        self._arrayDef = JSGArrayExpr(self._context, ctx)

    def visitValueTypeMacro(self, ctx: jsgParser.ValueTypeMacroContext):
        # valueTypeMacro: ID EQUALS nonRefValueType (BAR nonRefValueType)* SEMI;
        if len(ctx.nonRefValueType()) > 1:
            self._proc_value_types(ctx.nonRefValueType())
        else:
            self.visit(ctx.nonRefValueType(0))

    def visitTypeAlternatives(self, ctx: jsgParser.TypeAlternativesContext):
        self._proc_value_types(ctx.valueType())

    def _proc_value_types(self,
                          ctx: Union[List[jsgParser.ValueTypeContext],
                                     List[jsgParser.NonRefValueTypeContext]]):
        from pyjsg.parser_impl.jsg_lexerruleblock_parser import JSGLexerRuleBlock

        stringalts = []                 # Aggregate multiple strings into a single type
        for vt in ctx:
            nrvt = vt.nonRefValueType() if isinstance(vt, jsgParser.ValueTypeContext) \
                else cast(jsgParser.NonRefValueTypeContext, vt)
            stringval = nrvt.STRING() if nrvt else None
            if stringval:
                stringalts.append(stringval.getText()[1:-1])
            else:
                self._alttypelist.append(JSGValueType(self._context, vt))

        if stringalts:
            lrb = JSGLexerRuleBlock(self._context)
            lrb.add_strings(stringalts)
            self._lexeridref = self._context.anon_id()
            self._context.grammarelts[self._lexeridref] = lrb



from typing import Optional, Dict, Any, Type, Union, TextIO

from pyjsg.jsglib.jsg_object import JSGObject
from pyjsg.jsglib.conformance import conforms
from pyjsg.jsglib.jsg_strings import JSGString
from pyjsg.jsglib.jsg_validateable import JSGValidateable
from pyjsg.jsglib.jsg_anytype import AnyType
from pyjsg.jsglib.jsg_null import JSGNull

from pyjsg.jsglib.logger import Logger


class JSGObjectMap(JSGObject):
    """
    An object map is a JsonObj with constraints on the attribute names, value types of both
    """
    _name_filter: type(JSGString) = None
    _value_type: type(JSGValidateable) = AnyType

    _min: int = 0
    _max: Optional[int] = None

    _strict = False

    def __setattr__(self, key: str, value: Any):
        """
        Screen attributes for name and type.  Anything starting with underscore ('_') goes, anything in the IGNORE list
        and anything declared in the __init_ signature
        :param key:
        :param value:
        :return:
        """

        if not key.startswith("_") and not self._context.unvalidated_parm(key):
            if self._name_filter is not None:
                if not isinstance(key, self._name_filter):
                    raise ValueError(f"Illegal Object Map key: {key}={value}")
            if not conforms(value, self._value_type, self._context.NAMESPACE):
                raise ValueError("Illegal value type {} = {}".format(key, value))
            if not isinstance(value, JSGValidateable):
                value = self._value_type('', self._context, value) \
                    if self._value_type is AnyType else self._value_type(value)
        self[key] = value

    def __getattribute__(self, item) -> Any:
        rval = super().__getattribute__(item)
        return rval.val if type(rval) is AnyType else None if rval is JSGNull else rval

    def __getitem__(self, item):
        rval = super().__getitem__(item)
        return rval.val if type(rval) is AnyType else None if rval is JSGNull else rval

    def _is_valid_element(self, log: Logger, name: str, entry: Type[JSGValidateable]) -> bool:
        if self._name_filter is not None:
            if not self._name_filter.matches(name):
                if log.log(f"Illegal Object Map key: {name}={entry}"):
                    return False
        return super()._is_valid_element(log, name, entry)

    def _is_valid(self, log_file: Optional[Union[Logger, TextIO]] = None) -> bool:
        log = Logger() if log_file is None else log_file if isinstance(log_file, Logger) else Logger(log_file)
        nerrors = log.nerrors
        nitems = len(self._strip_nones(self.__dict__))
        if nitems < self._min:
            if log.log(f"Number of elements is {nitems} which is less than the minimum number ({self._min})"):
                return False
        if self._max is not None and nitems > self._max:
            if log.log(f"Number of elements is {nitems} which is greater than the minimum number ({self._max})"):
                return False
        super()._is_valid(log)
        return nerrors == log.nerrors


from typing import List, Dict, Any


class JSGContext:
    """ Context available to all JSG constructs """
    def __init__(self):
        # the object member name, if any, that identifies the type of object. If present, must match the name of
        # a JSGObject
        self.TYPE: str = ""

        # Objects that lack type identifiers.  Any objects lacking a TYPE variable will be matched against
        # the list below in order
        self.TYPE_EXCEPTIONS: List[str] = []

        # Object pair names that can always exist in an object
        self.IGNORE: List[str] = []

        # True means that we allow JSON_LD constructs (parameters starting with "@")
        self.JSON_LD = True

        # NAMESPACE prevents references from being resolved against other JSG modules
        self.NAMESPACE: Dict[str, Any] = None

    def unvalidated_parm(self, parm: str) -> bool:
        """Return true if the pair name should be ignored

        :param parm: string part of pair string:value
        :return: True if it should be accepted
        """
        return parm.startswith("_") or parm == self.TYPE or parm in self.IGNORE or \
            (self.JSON_LD and parm.startswith('@'))



""" PyJSG -- Python JSON Schema Grammar Bindings

This package translates between JSON and Python objects, following the rules
specified by {JSON Schema Grammar<http://github.com/ericprud/jsglib>}

"""

__version__ = '0.10.0'
__url__ = 'http://github.com/hsolbrig/pyjsg'
__license__ = 'CC0 1.0 Universal'





class Empty:
    """ An empty element. Serves the same function as None, except that it can be used to indicate that JSGNull
    and AnyType attributes (which both can legitimately have None values) have not been assigned
    """
    def __new__(cls):
        return cls


from collections import OrderedDict
from typing import Optional, List, Tuple, Dict

from pyjsg.parser.jsgParser import *
from pyjsg.parser.jsgParserVisitor import jsgParserVisitor
from pyjsg.parser_impl.jsg_doc_context import JSGDocContext, PythonGeneratorElement
from pyjsg.parser_impl.jsg_ebnf_parser import JSGEbnf
from pyjsg.parser_impl.jsg_valuetype_parser import JSGValueType
from pyjsg.parser_impl.parser_utils import as_token, is_valid_python, get_terminal, esc_kw, flatten, t


class JSGPairDef(jsgParserVisitor, PythonGeneratorElement):
    def __init__(self, context: JSGDocContext, ctx: Optional[jsgParser.PairDefContext] = None):
        self._context = context

        # PairDef can be one of "name: valueType [ebnsuffix]", "(name [|] name ...): valueType [ebnsuffix]" or type_ref
        self._typ: Optional[JSGValueType] = None       # If absent, then _type reference
        self._names: Dict[str, str] = OrderedDict()    # List of names associated with _typ

        self._type_reference: Optional[str] = None     # Reference to external type

        self._ebnf = JSGEbnf(context)                  # Cardinality of either branch
        self.text = ""

        if ctx:
            self.text = ctx.getText()
            self.visit(ctx)

    def __str__(self):
        if self._typ:
            names = list(self._names.keys())
            if len(names) > 1:
                namelist = '(' + ' | '.join(name for name in names) + ')'
            else:
                namelist = names[0]
            return f"pairDef: {namelist} : {self._typ}{self._ebnf}"
        else:
            return f"pairDef: typeReference: {self._type_reference}{self._ebnf}"

    def is_reference_type(self) -> bool:
        return self._type_reference is not None

    def members_entries(self, all_are_optional: Optional[bool] = False) -> List[Tuple[str, str]]:
        """ Generate a list quoted raw name, signature type entries for this pairdef, recursively traversing
        reference types

        :param all_are_optional: If true, all types are forced optional
        :return: raw name/ signature type for all elements in this pair
        """
        if self._type_reference:
            rval: List[Tuple[str, str]] = []
            for n, t in self._context.reference(self._type_reference).members_entries(all_are_optional):
                rval.append((n, self._ebnf.signature_cardinality(t, all_are_optional).format(name=n)))
            return rval
        else:
            sig = self._ebnf.signature_cardinality(self._typ.reference_type(), all_are_optional)
            return [(name, sig.format(name=name)) for name in self._names]

    def signature_type(self) -> str:
        base_type = self._typ.signature_type() if self._typ \
            else self._context.reference(self._type_reference).signature_type()
        return self._ebnf.signature_cardinality(base_type)

    def reference_type(self) -> str:
        return self.signature_type()

    def python_base_type(self) -> str:
        return self._typ.python_type() if self._typ \
            else self._context.reference(self._type_reference).python_type()

    def python_type(self) -> str:
        return self._ebnf.python_cardinality(self.python_base_type())

    def mt_value(self) -> str:
        return self._typ.mt_value() if self._typ else self._context.reference(self._type_reference).mt_value()

    def signatures(self, all_are_optional: Optional[bool] = False) -> List[str]:
        """ Return the __init__ signature element(s) (var: type = default value).  Note that signatures are not
        generated for non-python names, although we do take the liberty of suffixing a '_' for reserved words
        (e.g. class: @int  generates "class_: int = None"

        :param all_are_optional: If True, all items are considered to be optional

        :return: List of signatures
        """
        if self._type_reference:
            # This assumes that references are to things that have signatures
            ref = self._context.reference(self._type_reference)
            if not getattr(ref, 'signatures', None):
                raise NotImplementedError("Reference to " + self._type_reference + " is not valid")
            return self._context.reference(self._type_reference).signatures(all_are_optional)
        else:
            return [f"{self._names[rn]}: {self.python_type()} = " 
                    f"{self._ebnf.mt_value(self._typ)}" for rn, cn in self._names.items() if is_valid_python(cn)]

    def _initializer_for(self, raw_name: str, cooked_name: str, prefix: Optional[str]) -> List[str]:
        """Create an initializer entry for the entry

        :param raw_name: name unadjusted for python compatibility.
        :param cooked_name: name that may or may not be python compatible

        :param prefix: owner of the element - used when objects passed as arguments

        :return: Initialization statements
        """
        mt_val = self._ebnf.mt_value(self._typ)
        rval = []

        if is_valid_python(raw_name):
            if prefix:
                # If a prefix exists, the input has already been processed - no if clause is necessary
                rval.append(f"self.{raw_name} = {prefix}.{raw_name}")
            else:
                cons = raw_name
                rval.append(f"self.{raw_name} = {cons}")
        elif is_valid_python(cooked_name):
            if prefix:
                rval.append(f"setattr(self, '{raw_name}', getattr({prefix}, '{raw_name}')")
            else:
                cons = f"{cooked_name} if {cooked_name} is not {mt_val} else _kwargs.get('{raw_name}', {mt_val})"
                rval.append(f"setattr(self, '{raw_name}', {cons})")
        else:
            getter = f"_kwargs.get('{raw_name}', {mt_val})"
            if prefix:
                rval.append(f"setattr(self, '{raw_name}', getattr({prefix}, '{getter}')")
            else:
                rval.append(f"setattr(self, '{raw_name}', {getter})")

        return rval

    def initializers(self, prefix: Optional[str] = None) -> List[str]:
        """ Return the __init__ initializer assignment block """
        if self._type_reference:
            # This assumes that references are to things that have initializers
            # TODO: Remove this check once we are certian things are good
            ref = self._context.reference(self._type_reference)
            if not getattr(ref, 'signatures', None):
                raise NotImplementedError("Reference to " + self._type_reference + " is not valid")
            return self._context.reference(self._type_reference).initializers(prefix)
        else:
            return flatten([self._initializer_for(rn, cn, prefix) for rn, cn in self._names.items()])

    def dependency_list(self) -> List[str]:
        return self._typ.dependency_list() if self._typ else [self._type_reference]

    # ***************
    #   Visitors
    # ***************
    def visitPairDef(self, ctx: jsgParser.PairDefContext):
        """ pairDef: name COLON valueType ebnfSuffix? 
                     | idref ebnfSuffix?
                     | OPREN name (BAR? name)+ CPREN COLON valueType ebnfSuffix?
        """
        if ctx.name():          # Options 1 or 3
            self.visitChildren(ctx)
        else:
            self._type_reference = as_token(ctx)
            if ctx.ebnfSuffix():
                self.visit(ctx.ebnfSuffix())

    def visitName(self, ctx: jsgParser.NameContext):
        """ name: ID | STRING """
        rtkn = get_terminal(ctx)
        tkn = esc_kw(rtkn)
        self._names[rtkn] = tkn

    def visitValueType(self, ctx: jsgParser.ValueTypeContext):
        self._typ = JSGValueType(self._context, ctx)

    def visitEbnfSuffix(self, ctx: jsgParser.EbnfSuffixContext):
        self._ebnf = JSGEbnf(self._context, ctx)



# Generated from /Users/mrf7578/Development/git/hsolbrig/pyjsg/grammar/jsgParser.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\35")
        buf.write("\u0131\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
        buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
        buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\3\2\7\2")
        buf.write("<\n\2\f\2\16\2?\13\2\3\2\7\2B\n\2\f\2\16\2E\13\2\3\2\3")
        buf.write("\2\7\2I\n\2\f\2\16\2L\13\2\3\2\3\2\3\3\3\3\3\3\3\3\5\3")
        buf.write("T\n\3\3\3\3\3\3\3\7\3Y\n\3\f\3\16\3\\\13\3\3\3\5\3_\n")
        buf.write("\3\3\4\6\4b\n\4\r\4\16\4c\3\5\3\5\3\5\3\5\5\5j\n\5\3\6")
        buf.write("\3\6\3\6\3\7\3\7\5\7q\n\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7")
        buf.write("\5\7z\n\7\3\b\6\b}\n\b\r\b\16\b~\3\b\3\b\7\b\u0083\n\b")
        buf.write("\f\b\16\b\u0086\13\b\3\t\7\t\u0089\n\t\f\t\16\t\u008c")
        buf.write("\13\t\3\n\3\n\3\n\3\13\3\13\3\13\3\13\7\13\u0095\n\13")
        buf.write("\f\13\16\13\u0098\13\13\3\13\5\13\u009b\n\13\3\13\3\13")
        buf.write("\3\f\3\f\5\f\u00a1\n\f\3\f\5\f\u00a4\n\f\3\f\3\f\5\f\u00a8")
        buf.write("\n\f\5\f\u00aa\n\f\3\r\3\r\3\r\3\r\5\r\u00b0\n\r\3\r\3")
        buf.write("\r\3\r\3\r\6\r\u00b6\n\r\r\r\16\r\u00b7\3\r\3\r\3\r\3")
        buf.write("\r\5\r\u00be\n\r\3\r\3\r\3\r\3\r\6\r\u00c4\n\r\r\r\16")
        buf.write("\r\u00c5\3\r\3\r\5\r\u00ca\n\r\3\16\6\16\u00cd\n\16\r")
        buf.write("\16\16\16\u00ce\3\17\3\17\3\17\3\17\3\17\3\17\3\17\3\17")
        buf.write("\3\17\3\17\5\17\u00db\n\17\3\20\3\20\3\20\6\20\u00e0\n")
        buf.write("\20\r\20\16\20\u00e1\3\21\3\21\3\21\3\21\3\21\3\22\3\22")
        buf.write("\3\22\3\22\3\22\3\23\3\23\3\23\7\23\u00f1\n\23\f\23\16")
        buf.write("\23\u00f4\13\23\3\24\3\24\3\24\3\24\3\24\3\24\3\24\5\24")
        buf.write("\u00fd\n\24\5\24\u00ff\n\24\3\24\5\24\u0102\n\24\3\25")
        buf.write("\3\25\3\25\3\25\3\25\3\26\3\26\3\27\3\27\3\27\7\27\u010e")
        buf.write("\n\27\f\27\16\27\u0111\13\27\3\30\3\30\5\30\u0115\n\30")
        buf.write("\3\31\6\31\u0118\n\31\r\31\16\31\u0119\3\32\3\32\5\32")
        buf.write("\u011e\n\32\3\32\3\32\5\32\u0122\n\32\5\32\u0124\n\32")
        buf.write("\3\33\3\33\3\33\3\33\3\34\3\34\3\34\5\34\u012d\n\34\3")
        buf.write("\35\3\35\3\35\2\2\36\2\4\6\b\n\f\16\20\22\24\26\30\32")
        buf.write("\34\36 \"$&(*,.\60\62\64\668\2\6\3\2\27\30\3\2\26\30\4")
        buf.write("\2\24\24\31\31\4\2\26\26\30\30\2\u0145\2=\3\2\2\2\4^\3")
        buf.write("\2\2\2\6a\3\2\2\2\bi\3\2\2\2\nk\3\2\2\2\fy\3\2\2\2\16")
        buf.write("|\3\2\2\2\20\u008a\3\2\2\2\22\u008d\3\2\2\2\24\u0090\3")
        buf.write("\2\2\2\26\u00a9\3\2\2\2\30\u00c9\3\2\2\2\32\u00cc\3\2")
        buf.write("\2\2\34\u00da\3\2\2\2\36\u00dc\3\2\2\2 \u00e3\3\2\2\2")
        buf.write("\"\u00e8\3\2\2\2$\u00ed\3\2\2\2&\u0101\3\2\2\2(\u0103")
        buf.write("\3\2\2\2*\u0108\3\2\2\2,\u010a\3\2\2\2.\u0114\3\2\2\2")
        buf.write("\60\u0117\3\2\2\2\62\u0123\3\2\2\2\64\u0125\3\2\2\2\66")
        buf.write("\u012c\3\2\2\28\u012e\3\2\2\2:<\5\4\3\2;:\3\2\2\2<?\3")
        buf.write("\2\2\2=;\3\2\2\2=>\3\2\2\2>C\3\2\2\2?=\3\2\2\2@B\5\b\5")
        buf.write("\2A@\3\2\2\2BE\3\2\2\2CA\3\2\2\2CD\3\2\2\2DF\3\2\2\2E")
        buf.write("C\3\2\2\2FJ\7\3\2\2GI\5(\25\2HG\3\2\2\2IL\3\2\2\2JH\3")
        buf.write("\2\2\2JK\3\2\2\2KM\3\2\2\2LJ\3\2\2\2MN\7\2\2\3N\3\3\2")
        buf.write("\2\2OP\7\4\2\2PS\7\27\2\2QR\7\5\2\2RT\5\6\4\2SQ\3\2\2")
        buf.write("\2ST\3\2\2\2TU\3\2\2\2U_\7\6\2\2VZ\7\7\2\2WY\7\27\2\2")
        buf.write("XW\3\2\2\2Y\\\3\2\2\2ZX\3\2\2\2Z[\3\2\2\2[]\3\2\2\2\\")
        buf.write("Z\3\2\2\2]_\7\6\2\2^O\3\2\2\2^V\3\2\2\2_\5\3\2\2\2`b\7")
        buf.write("\27\2\2a`\3\2\2\2bc\3\2\2\2ca\3\2\2\2cd\3\2\2\2d\7\3\2")
        buf.write("\2\2ej\5\n\6\2fj\5\22\n\2gj\5 \21\2hj\5\"\22\2ie\3\2\2")
        buf.write("\2if\3\2\2\2ig\3\2\2\2ih\3\2\2\2j\t\3\2\2\2kl\7\27\2\2")
        buf.write("lm\5\f\7\2m\13\3\2\2\2np\7\b\2\2oq\5\16\b\2po\3\2\2\2")
        buf.write("pq\3\2\2\2qr\3\2\2\2rz\7\t\2\2st\7\b\2\2tu\7\27\2\2uv")
        buf.write("\7\n\2\2vw\5\34\17\2wx\7\t\2\2xz\3\2\2\2yn\3\2\2\2ys\3")
        buf.write("\2\2\2z\r\3\2\2\2{}\5\26\f\2|{\3\2\2\2}~\3\2\2\2~|\3\2")
        buf.write("\2\2~\177\3\2\2\2\177\u0084\3\2\2\2\u0080\u0081\7\13\2")
        buf.write("\2\u0081\u0083\5\20\t\2\u0082\u0080\3\2\2\2\u0083\u0086")
        buf.write("\3\2\2\2\u0084\u0082\3\2\2\2\u0084\u0085\3\2\2\2\u0085")
        buf.write("\17\3\2\2\2\u0086\u0084\3\2\2\2\u0087\u0089\5\26\f\2\u0088")
        buf.write("\u0087\3\2\2\2\u0089\u008c\3\2\2\2\u008a\u0088\3\2\2\2")
        buf.write("\u008a\u008b\3\2\2\2\u008b\21\3\2\2\2\u008c\u008a\3\2")
        buf.write("\2\2\u008d\u008e\7\27\2\2\u008e\u008f\5\24\13\2\u008f")
        buf.write("\23\3\2\2\2\u0090\u0091\7\f\2\2\u0091\u0096\5\34\17\2")
        buf.write("\u0092\u0093\7\13\2\2\u0093\u0095\5\34\17\2\u0094\u0092")
        buf.write("\3\2\2\2\u0095\u0098\3\2\2\2\u0096\u0094\3\2\2\2\u0096")
        buf.write("\u0097\3\2\2\2\u0097\u009a\3\2\2\2\u0098\u0096\3\2\2\2")
        buf.write("\u0099\u009b\5&\24\2\u009a\u0099\3\2\2\2\u009a\u009b\3")
        buf.write("\2\2\2\u009b\u009c\3\2\2\2\u009c\u009d\7\r\2\2\u009d\25")
        buf.write("\3\2\2\2\u009e\u00a0\7\27\2\2\u009f\u00a1\5&\24\2\u00a0")
        buf.write("\u009f\3\2\2\2\u00a0\u00a1\3\2\2\2\u00a1\u00a3\3\2\2\2")
        buf.write("\u00a2\u00a4\7\16\2\2\u00a3\u00a2\3\2\2\2\u00a3\u00a4")
        buf.write("\3\2\2\2\u00a4\u00aa\3\2\2\2\u00a5\u00a7\5\30\r\2\u00a6")
        buf.write("\u00a8\7\16\2\2\u00a7\u00a6\3\2\2\2\u00a7\u00a8\3\2\2")
        buf.write("\2\u00a8\u00aa\3\2\2\2\u00a9\u009e\3\2\2\2\u00a9\u00a5")
        buf.write("\3\2\2\2\u00aa\27\3\2\2\2\u00ab\u00ac\t\2\2\2\u00ac\u00ad")
        buf.write("\7\17\2\2\u00ad\u00af\5\34\17\2\u00ae\u00b0\5&\24\2\u00af")
        buf.write("\u00ae\3\2\2\2\u00af\u00b0\3\2\2\2\u00b0\u00ca\3\2\2\2")
        buf.write("\u00b1\u00b2\7\20\2\2\u00b2\u00b5\7\27\2\2\u00b3\u00b4")
        buf.write("\7\13\2\2\u00b4\u00b6\7\27\2\2\u00b5\u00b3\3\2\2\2\u00b6")
        buf.write("\u00b7\3\2\2\2\u00b7\u00b5\3\2\2\2\u00b7\u00b8\3\2\2\2")
        buf.write("\u00b8\u00b9\3\2\2\2\u00b9\u00ba\7\21\2\2\u00ba\u00bb")
        buf.write("\7\17\2\2\u00bb\u00bd\5\34\17\2\u00bc\u00be\5&\24\2\u00bd")
        buf.write("\u00bc\3\2\2\2\u00bd\u00be\3\2\2\2\u00be\u00ca\3\2\2\2")
        buf.write("\u00bf\u00c0\7\20\2\2\u00c0\u00c3\5\32\16\2\u00c1\u00c2")
        buf.write("\7\13\2\2\u00c2\u00c4\5\32\16\2\u00c3\u00c1\3\2\2\2\u00c4")
        buf.write("\u00c5\3\2\2\2\u00c5\u00c3\3\2\2\2\u00c5\u00c6\3\2\2\2")
        buf.write("\u00c6\u00c7\3\2\2\2\u00c7\u00c8\7\21\2\2\u00c8\u00ca")
        buf.write("\3\2\2\2\u00c9\u00ab\3\2\2\2\u00c9\u00b1\3\2\2\2\u00c9")
        buf.write("\u00bf\3\2\2\2\u00ca\31\3\2\2\2\u00cb\u00cd\5\30\r\2\u00cc")
        buf.write("\u00cb\3\2\2\2\u00cd\u00ce\3\2\2\2\u00ce\u00cc\3\2\2\2")
        buf.write("\u00ce\u00cf\3\2\2\2\u00cf\33\3\2\2\2\u00d0\u00db\7\27")
        buf.write("\2\2\u00d1\u00db\7\26\2\2\u00d2\u00db\7\30\2\2\u00d3\u00db")
        buf.write("\5\f\7\2\u00d4\u00db\5\24\13\2\u00d5\u00d6\7\20\2\2\u00d6")
        buf.write("\u00d7\5\36\20\2\u00d7\u00d8\7\21\2\2\u00d8\u00db\3\2")
        buf.write("\2\2\u00d9\u00db\7\32\2\2\u00da\u00d0\3\2\2\2\u00da\u00d1")
        buf.write("\3\2\2\2\u00da\u00d2\3\2\2\2\u00da\u00d3\3\2\2\2\u00da")
        buf.write("\u00d4\3\2\2\2\u00da\u00d5\3\2\2\2\u00da\u00d9\3\2\2\2")
        buf.write("\u00db\35\3\2\2\2\u00dc\u00df\t\3\2\2\u00dd\u00de\7\13")
        buf.write("\2\2\u00de\u00e0\t\3\2\2\u00df\u00dd\3\2\2\2\u00e0\u00e1")
        buf.write("\3\2\2\2\u00e1\u00df\3\2\2\2\u00e1\u00e2\3\2\2\2\u00e2")
        buf.write("\37\3\2\2\2\u00e3\u00e4\7\27\2\2\u00e4\u00e5\7\22\2\2")
        buf.write("\u00e5\u00e6\5\16\b\2\u00e6\u00e7\7\6\2\2\u00e7!\3\2\2")
        buf.write("\2\u00e8\u00e9\7\27\2\2\u00e9\u00ea\7\22\2\2\u00ea\u00eb")
        buf.write("\5$\23\2\u00eb\u00ec\7\6\2\2\u00ec#\3\2\2\2\u00ed\u00f2")
        buf.write("\7\26\2\2\u00ee\u00ef\7\13\2\2\u00ef\u00f1\7\26\2\2\u00f0")
        buf.write("\u00ee\3\2\2\2\u00f1\u00f4\3\2\2\2\u00f2\u00f0\3\2\2\2")
        buf.write("\u00f2\u00f3\3\2\2\2\u00f3%\3\2\2\2\u00f4\u00f2\3\2\2")
        buf.write("\2\u00f5\u0102\7\23\2\2\u00f6\u0102\7\24\2\2\u00f7\u0102")
        buf.write("\7\25\2\2\u00f8\u00f9\7\b\2\2\u00f9\u00fe\7\31\2\2\u00fa")
        buf.write("\u00fc\7\16\2\2\u00fb\u00fd\t\4\2\2\u00fc\u00fb\3\2\2")
        buf.write("\2\u00fc\u00fd\3\2\2\2\u00fd\u00ff\3\2\2\2\u00fe\u00fa")
        buf.write("\3\2\2\2\u00fe\u00ff\3\2\2\2\u00ff\u0100\3\2\2\2\u0100")
        buf.write("\u0102\7\t\2\2\u0101\u00f5\3\2\2\2\u0101\u00f6\3\2\2\2")
        buf.write("\u0101\u00f7\3\2\2\2\u0101\u00f8\3\2\2\2\u0102\'\3\2\2")
        buf.write("\2\u0103\u0104\7\26\2\2\u0104\u0105\7\17\2\2\u0105\u0106")
        buf.write("\5*\26\2\u0106\u0107\7\6\2\2\u0107)\3\2\2\2\u0108\u0109")
        buf.write("\5,\27\2\u0109+\3\2\2\2\u010a\u010f\5.\30\2\u010b\u010c")
        buf.write("\7\13\2\2\u010c\u010e\5.\30\2\u010d\u010b\3\2\2\2\u010e")
        buf.write("\u0111\3\2\2\2\u010f\u010d\3\2\2\2\u010f\u0110\3\2\2\2")
        buf.write("\u0110-\3\2\2\2\u0111\u010f\3\2\2\2\u0112\u0115\5\60\31")
        buf.write("\2\u0113\u0115\3\2\2\2\u0114\u0112\3\2\2\2\u0114\u0113")
        buf.write("\3\2\2\2\u0115/\3\2\2\2\u0116\u0118\5\62\32\2\u0117\u0116")
        buf.write("\3\2\2\2\u0118\u0119\3\2\2\2\u0119\u0117\3\2\2\2\u0119")
        buf.write("\u011a\3\2\2\2\u011a\61\3\2\2\2\u011b\u011d\5\66\34\2")
        buf.write("\u011c\u011e\5&\24\2\u011d\u011c\3\2\2\2\u011d\u011e\3")
        buf.write("\2\2\2\u011e\u0124\3\2\2\2\u011f\u0121\5\64\33\2\u0120")
        buf.write("\u0122\5&\24\2\u0121\u0120\3\2\2\2\u0121\u0122\3\2\2\2")
        buf.write("\u0122\u0124\3\2\2\2\u0123\u011b\3\2\2\2\u0123\u011f\3")
        buf.write("\2\2\2\u0124\63\3\2\2\2\u0125\u0126\7\20\2\2\u0126\u0127")
        buf.write("\5,\27\2\u0127\u0128\7\21\2\2\u0128\65\3\2\2\2\u0129\u012d")
        buf.write("\58\35\2\u012a\u012d\7\35\2\2\u012b\u012d\7\32\2\2\u012c")
        buf.write("\u0129\3\2\2\2\u012c\u012a\3\2\2\2\u012c\u012b\3\2\2\2")
        buf.write("\u012d\67\3\2\2\2\u012e\u012f\t\5\2\2\u012f9\3\2\2\2(")
        buf.write("=CJSZ^cipy~\u0084\u008a\u0096\u009a\u00a0\u00a3\u00a7")
        buf.write("\u00a9\u00af\u00b7\u00bd\u00c5\u00c9\u00ce\u00da\u00e1")
        buf.write("\u00f2\u00fc\u00fe\u0101\u010f\u0114\u0119\u011d\u0121")
        buf.write("\u0123\u012c")
        return buf.getvalue()


class jsgParserParser ( Parser ):

    grammarFileName = "jsgParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'@terminals'", "'.TYPE'", "'-'", "';'", 
                     "'.IGNORE'", "'{'", "'}'", "'->'", "'|'", "'['", "']'", 
                     "','", "':'", "'('", "')'", "'='", "'?'", "'*'", "'+'", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'.'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "LEXER_ID", "ID", "STRING", "INT", "ANY", "PASS", 
                      "COMMENT", "LEXER_CHAR_SET" ]

    RULE_doc = 0
    RULE_directive = 1
    RULE_typeExceptions = 2
    RULE_grammarElt = 3
    RULE_objectDef = 4
    RULE_objectExpr = 5
    RULE_objectExprDef = 6
    RULE_particleOpt = 7
    RULE_arrayDef = 8
    RULE_arrayExpr = 9
    RULE_particle = 10
    RULE_propertyOrGroup = 11
    RULE_propertyOrGroupList = 12
    RULE_propertyType = 13
    RULE_typeAlternatives = 14
    RULE_objectAlias = 15
    RULE_lexerAlias = 16
    RULE_lexerAliasDef = 17
    RULE_ebnfSuffix = 18
    RULE_lexerRuleSpec = 19
    RULE_lexerRuleBlock = 20
    RULE_lexerAltList = 21
    RULE_lexerAlt = 22
    RULE_lexerElements = 23
    RULE_lexerElement = 24
    RULE_lexerBlock = 25
    RULE_lexerAtom = 26
    RULE_terminal = 27

    ruleNames =  [ "doc", "directive", "typeExceptions", "grammarElt", "objectDef", 
                   "objectExpr", "objectExprDef", "particleOpt", "arrayDef", 
                   "arrayExpr", "particle", "propertyOrGroup", "propertyOrGroupList", 
                   "propertyType", "typeAlternatives", "objectAlias", "lexerAlias", 
                   "lexerAliasDef", "ebnfSuffix", "lexerRuleSpec", "lexerRuleBlock", 
                   "lexerAltList", "lexerAlt", "lexerElements", "lexerElement", 
                   "lexerBlock", "lexerAtom", "terminal" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    LEXER_ID=20
    ID=21
    STRING=22
    INT=23
    ANY=24
    PASS=25
    COMMENT=26
    LEXER_CHAR_SET=27

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class DocContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(jsgParserParser.EOF, 0)

        def directive(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.DirectiveContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.DirectiveContext,i)


        def grammarElt(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.GrammarEltContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.GrammarEltContext,i)


        def lexerRuleSpec(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.LexerRuleSpecContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.LexerRuleSpecContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_doc

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDoc" ):
                return visitor.visitDoc(self)
            else:
                return visitor.visitChildren(self)




    def doc(self):

        localctx = jsgParserParser.DocContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_doc)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 59
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.T__1 or _la==jsgParserParser.T__4:
                self.state = 56
                self.directive()
                self.state = 61
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 65
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.ID:
                self.state = 62
                self.grammarElt()
                self.state = 67
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 68
            self.match(jsgParserParser.T__0)
            self.state = 72
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.LEXER_ID:
                self.state = 69
                self.lexerRuleSpec()
                self.state = 74
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 75
            self.match(jsgParserParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DirectiveContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return jsgParserParser.RULE_directive

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class TypeDirectiveContext(DirectiveContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.DirectiveContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)
        def typeExceptions(self):
            return self.getTypedRuleContext(jsgParserParser.TypeExceptionsContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTypeDirective" ):
                return visitor.visitTypeDirective(self)
            else:
                return visitor.visitChildren(self)


    class IgnoreDirectiveContext(DirectiveContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.DirectiveContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.ID)
            else:
                return self.getToken(jsgParserParser.ID, i)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIgnoreDirective" ):
                return visitor.visitIgnoreDirective(self)
            else:
                return visitor.visitChildren(self)



    def directive(self):

        localctx = jsgParserParser.DirectiveContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_directive)
        self._la = 0 # Token type
        try:
            self.state = 92
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.T__1]:
                localctx = jsgParserParser.TypeDirectiveContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 77
                self.match(jsgParserParser.T__1)
                self.state = 78
                self.match(jsgParserParser.ID)
                self.state = 81
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==jsgParserParser.T__2:
                    self.state = 79
                    self.match(jsgParserParser.T__2)
                    self.state = 80
                    self.typeExceptions()


                self.state = 83
                self.match(jsgParserParser.T__3)
                pass
            elif token in [jsgParserParser.T__4]:
                localctx = jsgParserParser.IgnoreDirectiveContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 84
                self.match(jsgParserParser.T__4)
                self.state = 88
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==jsgParserParser.ID:
                    self.state = 85
                    self.match(jsgParserParser.ID)
                    self.state = 90
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 91
                self.match(jsgParserParser.T__3)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TypeExceptionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.ID)
            else:
                return self.getToken(jsgParserParser.ID, i)

        def getRuleIndex(self):
            return jsgParserParser.RULE_typeExceptions

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTypeExceptions" ):
                return visitor.visitTypeExceptions(self)
            else:
                return visitor.visitChildren(self)




    def typeExceptions(self):

        localctx = jsgParserParser.TypeExceptionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_typeExceptions)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 94
                self.match(jsgParserParser.ID)
                self.state = 97 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==jsgParserParser.ID):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class GrammarEltContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def objectDef(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectDefContext,0)


        def arrayDef(self):
            return self.getTypedRuleContext(jsgParserParser.ArrayDefContext,0)


        def objectAlias(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectAliasContext,0)


        def lexerAlias(self):
            return self.getTypedRuleContext(jsgParserParser.LexerAliasContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_grammarElt

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitGrammarElt" ):
                return visitor.visitGrammarElt(self)
            else:
                return visitor.visitChildren(self)




    def grammarElt(self):

        localctx = jsgParserParser.GrammarEltContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_grammarElt)
        try:
            self.state = 103
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,7,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 99
                self.objectDef()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 100
                self.arrayDef()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 101
                self.objectAlias()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 102
                self.lexerAlias()
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectDefContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def objectExpr(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectExprContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_objectDef

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectDef" ):
                return visitor.visitObjectDef(self)
            else:
                return visitor.visitChildren(self)




    def objectDef(self):

        localctx = jsgParserParser.ObjectDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_objectDef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 105
            self.match(jsgParserParser.ID)
            self.state = 106
            self.objectExpr()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return jsgParserParser.RULE_objectExpr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ObjectExprMapContext(ObjectExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.ObjectExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)
        def propertyType(self):
            return self.getTypedRuleContext(jsgParserParser.PropertyTypeContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectExprMap" ):
                return visitor.visitObjectExprMap(self)
            else:
                return visitor.visitChildren(self)


    class ObjectExprObjContext(ObjectExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.ObjectExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectExprDef(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectExprDefContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectExprObj" ):
                return visitor.visitObjectExprObj(self)
            else:
                return visitor.visitChildren(self)



    def objectExpr(self):

        localctx = jsgParserParser.ObjectExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_objectExpr)
        self._la = 0 # Token type
        try:
            self.state = 119
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                localctx = jsgParserParser.ObjectExprObjContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 108
                self.match(jsgParserParser.T__5)
                self.state = 110
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__13) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0):
                    self.state = 109
                    self.objectExprDef()


                self.state = 112
                self.match(jsgParserParser.T__6)
                pass

            elif la_ == 2:
                localctx = jsgParserParser.ObjectExprMapContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 113
                self.match(jsgParserParser.T__5)
                self.state = 114
                self.match(jsgParserParser.ID)
                self.state = 115
                self.match(jsgParserParser.T__7)
                self.state = 116
                self.propertyType()
                self.state = 117
                self.match(jsgParserParser.T__6)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectExprDefContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def particle(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.ParticleContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.ParticleContext,i)


        def particleOpt(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.ParticleOptContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.ParticleOptContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_objectExprDef

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectExprDef" ):
                return visitor.visitObjectExprDef(self)
            else:
                return visitor.visitChildren(self)




    def objectExprDef(self):

        localctx = jsgParserParser.ObjectExprDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_objectExprDef)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 122 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 121
                self.particle()
                self.state = 124 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__13) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0)):
                    break

            self.state = 130
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.T__8:
                self.state = 126
                self.match(jsgParserParser.T__8)
                self.state = 127
                self.particleOpt()
                self.state = 132
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParticleOptContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def particle(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.ParticleContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.ParticleContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_particleOpt

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParticleOpt" ):
                return visitor.visitParticleOpt(self)
            else:
                return visitor.visitChildren(self)




    def particleOpt(self):

        localctx = jsgParserParser.ParticleOptContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_particleOpt)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 136
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__13) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0):
                self.state = 133
                self.particle()
                self.state = 138
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArrayDefContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def arrayExpr(self):
            return self.getTypedRuleContext(jsgParserParser.ArrayExprContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_arrayDef

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArrayDef" ):
                return visitor.visitArrayDef(self)
            else:
                return visitor.visitChildren(self)




    def arrayDef(self):

        localctx = jsgParserParser.ArrayDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_arrayDef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 139
            self.match(jsgParserParser.ID)
            self.state = 140
            self.arrayExpr()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArrayExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def propertyType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.PropertyTypeContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.PropertyTypeContext,i)


        def ebnfSuffix(self):
            return self.getTypedRuleContext(jsgParserParser.EbnfSuffixContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_arrayExpr

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArrayExpr" ):
                return visitor.visitArrayExpr(self)
            else:
                return visitor.visitChildren(self)




    def arrayExpr(self):

        localctx = jsgParserParser.ArrayExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_arrayExpr)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 142
            self.match(jsgParserParser.T__9)
            self.state = 143
            self.propertyType()
            self.state = 148
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.T__8:
                self.state = 144
                self.match(jsgParserParser.T__8)
                self.state = 145
                self.propertyType()
                self.state = 150
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 152
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                self.state = 151
                self.ebnfSuffix()


            self.state = 154
            self.match(jsgParserParser.T__10)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParticleContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def ebnfSuffix(self):
            return self.getTypedRuleContext(jsgParserParser.EbnfSuffixContext,0)


        def propertyOrGroup(self):
            return self.getTypedRuleContext(jsgParserParser.PropertyOrGroupContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_particle

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParticle" ):
                return visitor.visitParticle(self)
            else:
                return visitor.visitChildren(self)




    def particle(self):

        localctx = jsgParserParser.ParticleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_particle)
        self._la = 0 # Token type
        try:
            self.state = 167
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,18,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 156
                self.match(jsgParserParser.ID)
                self.state = 158
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                    self.state = 157
                    self.ebnfSuffix()


                self.state = 161
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==jsgParserParser.T__11:
                    self.state = 160
                    self.match(jsgParserParser.T__11)


                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 163
                self.propertyOrGroup()
                self.state = 165
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==jsgParserParser.T__11:
                    self.state = 164
                    self.match(jsgParserParser.T__11)


                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PropertyOrGroupContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return jsgParserParser.RULE_propertyOrGroup

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PropertyOrGroupChoiceContext(PropertyOrGroupContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.PropertyOrGroupContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def propertyOrGroupList(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.PropertyOrGroupListContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.PropertyOrGroupListContext,i)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropertyOrGroupChoice" ):
                return visitor.visitPropertyOrGroupChoice(self)
            else:
                return visitor.visitChildren(self)


    class PropertyOrGroupShorthandContext(PropertyOrGroupContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.PropertyOrGroupContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.ID)
            else:
                return self.getToken(jsgParserParser.ID, i)
        def propertyType(self):
            return self.getTypedRuleContext(jsgParserParser.PropertyTypeContext,0)

        def ebnfSuffix(self):
            return self.getTypedRuleContext(jsgParserParser.EbnfSuffixContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropertyOrGroupShorthand" ):
                return visitor.visitPropertyOrGroupShorthand(self)
            else:
                return visitor.visitChildren(self)


    class PropertyOrGroupSimpleContext(PropertyOrGroupContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a jsgParserParser.PropertyOrGroupContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def propertyType(self):
            return self.getTypedRuleContext(jsgParserParser.PropertyTypeContext,0)

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)
        def STRING(self):
            return self.getToken(jsgParserParser.STRING, 0)
        def ebnfSuffix(self):
            return self.getTypedRuleContext(jsgParserParser.EbnfSuffixContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropertyOrGroupSimple" ):
                return visitor.visitPropertyOrGroupSimple(self)
            else:
                return visitor.visitChildren(self)



    def propertyOrGroup(self):

        localctx = jsgParserParser.PropertyOrGroupContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_propertyOrGroup)
        self._la = 0 # Token type
        try:
            self.state = 199
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,23,self._ctx)
            if la_ == 1:
                localctx = jsgParserParser.PropertyOrGroupSimpleContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 169
                _la = self._input.LA(1)
                if not(_la==jsgParserParser.ID or _la==jsgParserParser.STRING):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 170
                self.match(jsgParserParser.T__12)
                self.state = 171
                self.propertyType()
                self.state = 173
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                    self.state = 172
                    self.ebnfSuffix()


                pass

            elif la_ == 2:
                localctx = jsgParserParser.PropertyOrGroupShorthandContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 175
                self.match(jsgParserParser.T__13)
                self.state = 176
                self.match(jsgParserParser.ID)
                self.state = 179 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 177
                    self.match(jsgParserParser.T__8)
                    self.state = 178
                    self.match(jsgParserParser.ID)
                    self.state = 181 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==jsgParserParser.T__8):
                        break

                self.state = 183
                self.match(jsgParserParser.T__14)
                self.state = 184
                self.match(jsgParserParser.T__12)
                self.state = 185
                self.propertyType()
                self.state = 187
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                    self.state = 186
                    self.ebnfSuffix()


                pass

            elif la_ == 3:
                localctx = jsgParserParser.PropertyOrGroupChoiceContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 189
                self.match(jsgParserParser.T__13)
                self.state = 190
                self.propertyOrGroupList()
                self.state = 193 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 191
                    self.match(jsgParserParser.T__8)
                    self.state = 192
                    self.propertyOrGroupList()
                    self.state = 195 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==jsgParserParser.T__8):
                        break

                self.state = 197
                self.match(jsgParserParser.T__14)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PropertyOrGroupListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def propertyOrGroup(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.PropertyOrGroupContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.PropertyOrGroupContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_propertyOrGroupList

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropertyOrGroupList" ):
                return visitor.visitPropertyOrGroupList(self)
            else:
                return visitor.visitChildren(self)




    def propertyOrGroupList(self):

        localctx = jsgParserParser.PropertyOrGroupListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_propertyOrGroupList)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 202 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 201
                self.propertyOrGroup()
                self.state = 204 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__13) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PropertyTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def LEXER_ID(self):
            return self.getToken(jsgParserParser.LEXER_ID, 0)

        def STRING(self):
            return self.getToken(jsgParserParser.STRING, 0)

        def objectExpr(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectExprContext,0)


        def arrayExpr(self):
            return self.getTypedRuleContext(jsgParserParser.ArrayExprContext,0)


        def typeAlternatives(self):
            return self.getTypedRuleContext(jsgParserParser.TypeAlternativesContext,0)


        def ANY(self):
            return self.getToken(jsgParserParser.ANY, 0)

        def getRuleIndex(self):
            return jsgParserParser.RULE_propertyType

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropertyType" ):
                return visitor.visitPropertyType(self)
            else:
                return visitor.visitChildren(self)




    def propertyType(self):

        localctx = jsgParserParser.PropertyTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_propertyType)
        try:
            self.state = 216
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.ID]:
                self.enterOuterAlt(localctx, 1)
                self.state = 206
                self.match(jsgParserParser.ID)
                pass
            elif token in [jsgParserParser.LEXER_ID]:
                self.enterOuterAlt(localctx, 2)
                self.state = 207
                self.match(jsgParserParser.LEXER_ID)
                pass
            elif token in [jsgParserParser.STRING]:
                self.enterOuterAlt(localctx, 3)
                self.state = 208
                self.match(jsgParserParser.STRING)
                pass
            elif token in [jsgParserParser.T__5]:
                self.enterOuterAlt(localctx, 4)
                self.state = 209
                self.objectExpr()
                pass
            elif token in [jsgParserParser.T__9]:
                self.enterOuterAlt(localctx, 5)
                self.state = 210
                self.arrayExpr()
                pass
            elif token in [jsgParserParser.T__13]:
                self.enterOuterAlt(localctx, 6)
                self.state = 211
                self.match(jsgParserParser.T__13)
                self.state = 212
                self.typeAlternatives()
                self.state = 213
                self.match(jsgParserParser.T__14)
                pass
            elif token in [jsgParserParser.ANY]:
                self.enterOuterAlt(localctx, 7)
                self.state = 215
                self.match(jsgParserParser.ANY)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TypeAlternativesContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LEXER_ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.LEXER_ID)
            else:
                return self.getToken(jsgParserParser.LEXER_ID, i)

        def ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.ID)
            else:
                return self.getToken(jsgParserParser.ID, i)

        def STRING(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.STRING)
            else:
                return self.getToken(jsgParserParser.STRING, i)

        def getRuleIndex(self):
            return jsgParserParser.RULE_typeAlternatives

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTypeAlternatives" ):
                return visitor.visitTypeAlternatives(self)
            else:
                return visitor.visitChildren(self)




    def typeAlternatives(self):

        localctx = jsgParserParser.TypeAlternativesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_typeAlternatives)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 218
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.LEXER_ID) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 221 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 219
                self.match(jsgParserParser.T__8)
                self.state = 220
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.LEXER_ID) | (1 << jsgParserParser.ID) | (1 << jsgParserParser.STRING))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 223 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==jsgParserParser.T__8):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectAliasContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def objectExprDef(self):
            return self.getTypedRuleContext(jsgParserParser.ObjectExprDefContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_objectAlias

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectAlias" ):
                return visitor.visitObjectAlias(self)
            else:
                return visitor.visitChildren(self)




    def objectAlias(self):

        localctx = jsgParserParser.ObjectAliasContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_objectAlias)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 225
            self.match(jsgParserParser.ID)
            self.state = 226
            self.match(jsgParserParser.T__15)
            self.state = 227
            self.objectExprDef()
            self.state = 228
            self.match(jsgParserParser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerAliasContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(jsgParserParser.ID, 0)

        def lexerAliasDef(self):
            return self.getTypedRuleContext(jsgParserParser.LexerAliasDefContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerAlias

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerAlias" ):
                return visitor.visitLexerAlias(self)
            else:
                return visitor.visitChildren(self)




    def lexerAlias(self):

        localctx = jsgParserParser.LexerAliasContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_lexerAlias)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 230
            self.match(jsgParserParser.ID)
            self.state = 231
            self.match(jsgParserParser.T__15)
            self.state = 232
            self.lexerAliasDef()
            self.state = 233
            self.match(jsgParserParser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerAliasDefContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LEXER_ID(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.LEXER_ID)
            else:
                return self.getToken(jsgParserParser.LEXER_ID, i)

        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerAliasDef

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerAliasDef" ):
                return visitor.visitLexerAliasDef(self)
            else:
                return visitor.visitChildren(self)




    def lexerAliasDef(self):

        localctx = jsgParserParser.LexerAliasDefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_lexerAliasDef)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 235
            self.match(jsgParserParser.LEXER_ID)
            self.state = 240
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.T__8:
                self.state = 236
                self.match(jsgParserParser.T__8)
                self.state = 237
                self.match(jsgParserParser.LEXER_ID)
                self.state = 242
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class EbnfSuffixContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self, i:int=None):
            if i is None:
                return self.getTokens(jsgParserParser.INT)
            else:
                return self.getToken(jsgParserParser.INT, i)

        def getRuleIndex(self):
            return jsgParserParser.RULE_ebnfSuffix

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitEbnfSuffix" ):
                return visitor.visitEbnfSuffix(self)
            else:
                return visitor.visitChildren(self)




    def ebnfSuffix(self):

        localctx = jsgParserParser.EbnfSuffixContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_ebnfSuffix)
        self._la = 0 # Token type
        try:
            self.state = 255
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.T__16]:
                self.enterOuterAlt(localctx, 1)
                self.state = 243
                self.match(jsgParserParser.T__16)
                pass
            elif token in [jsgParserParser.T__17]:
                self.enterOuterAlt(localctx, 2)
                self.state = 244
                self.match(jsgParserParser.T__17)
                pass
            elif token in [jsgParserParser.T__18]:
                self.enterOuterAlt(localctx, 3)
                self.state = 245
                self.match(jsgParserParser.T__18)
                pass
            elif token in [jsgParserParser.T__5]:
                self.enterOuterAlt(localctx, 4)
                self.state = 246
                self.match(jsgParserParser.T__5)
                self.state = 247
                self.match(jsgParserParser.INT)
                self.state = 252
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==jsgParserParser.T__11:
                    self.state = 248
                    self.match(jsgParserParser.T__11)
                    self.state = 250
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==jsgParserParser.T__17 or _la==jsgParserParser.INT:
                        self.state = 249
                        _la = self._input.LA(1)
                        if not(_la==jsgParserParser.T__17 or _la==jsgParserParser.INT):
                            self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()




                self.state = 254
                self.match(jsgParserParser.T__6)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerRuleSpecContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LEXER_ID(self):
            return self.getToken(jsgParserParser.LEXER_ID, 0)

        def lexerRuleBlock(self):
            return self.getTypedRuleContext(jsgParserParser.LexerRuleBlockContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerRuleSpec

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerRuleSpec" ):
                return visitor.visitLexerRuleSpec(self)
            else:
                return visitor.visitChildren(self)




    def lexerRuleSpec(self):

        localctx = jsgParserParser.LexerRuleSpecContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_lexerRuleSpec)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 257
            self.match(jsgParserParser.LEXER_ID)
            self.state = 258
            self.match(jsgParserParser.T__12)
            self.state = 259
            self.lexerRuleBlock()
            self.state = 260
            self.match(jsgParserParser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerRuleBlockContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerAltList(self):
            return self.getTypedRuleContext(jsgParserParser.LexerAltListContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerRuleBlock

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerRuleBlock" ):
                return visitor.visitLexerRuleBlock(self)
            else:
                return visitor.visitChildren(self)




    def lexerRuleBlock(self):

        localctx = jsgParserParser.LexerRuleBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_lexerRuleBlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 262
            self.lexerAltList()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerAltListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerAlt(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.LexerAltContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.LexerAltContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerAltList

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerAltList" ):
                return visitor.visitLexerAltList(self)
            else:
                return visitor.visitChildren(self)




    def lexerAltList(self):

        localctx = jsgParserParser.LexerAltListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_lexerAltList)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 264
            self.lexerAlt()
            self.state = 269
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==jsgParserParser.T__8:
                self.state = 265
                self.match(jsgParserParser.T__8)
                self.state = 266
                self.lexerAlt()
                self.state = 271
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerAltContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerElements(self):
            return self.getTypedRuleContext(jsgParserParser.LexerElementsContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerAlt

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerAlt" ):
                return visitor.visitLexerAlt(self)
            else:
                return visitor.visitChildren(self)




    def lexerAlt(self):

        localctx = jsgParserParser.LexerAltContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_lexerAlt)
        try:
            self.state = 274
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.T__13, jsgParserParser.LEXER_ID, jsgParserParser.STRING, jsgParserParser.ANY, jsgParserParser.LEXER_CHAR_SET]:
                self.enterOuterAlt(localctx, 1)
                self.state = 272
                self.lexerElements()
                pass
            elif token in [jsgParserParser.T__3, jsgParserParser.T__8, jsgParserParser.T__14]:
                self.enterOuterAlt(localctx, 2)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerElementsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerElement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(jsgParserParser.LexerElementContext)
            else:
                return self.getTypedRuleContext(jsgParserParser.LexerElementContext,i)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerElements

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerElements" ):
                return visitor.visitLexerElements(self)
            else:
                return visitor.visitChildren(self)




    def lexerElements(self):

        localctx = jsgParserParser.LexerElementsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_lexerElements)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 277 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 276
                self.lexerElement()
                self.state = 279 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__13) | (1 << jsgParserParser.LEXER_ID) | (1 << jsgParserParser.STRING) | (1 << jsgParserParser.ANY) | (1 << jsgParserParser.LEXER_CHAR_SET))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerElementContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerAtom(self):
            return self.getTypedRuleContext(jsgParserParser.LexerAtomContext,0)


        def ebnfSuffix(self):
            return self.getTypedRuleContext(jsgParserParser.EbnfSuffixContext,0)


        def lexerBlock(self):
            return self.getTypedRuleContext(jsgParserParser.LexerBlockContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerElement

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerElement" ):
                return visitor.visitLexerElement(self)
            else:
                return visitor.visitChildren(self)




    def lexerElement(self):

        localctx = jsgParserParser.LexerElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_lexerElement)
        self._la = 0 # Token type
        try:
            self.state = 289
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.LEXER_ID, jsgParserParser.STRING, jsgParserParser.ANY, jsgParserParser.LEXER_CHAR_SET]:
                self.enterOuterAlt(localctx, 1)
                self.state = 281
                self.lexerAtom()
                self.state = 283
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                    self.state = 282
                    self.ebnfSuffix()


                pass
            elif token in [jsgParserParser.T__13]:
                self.enterOuterAlt(localctx, 2)
                self.state = 285
                self.lexerBlock()
                self.state = 287
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << jsgParserParser.T__5) | (1 << jsgParserParser.T__16) | (1 << jsgParserParser.T__17) | (1 << jsgParserParser.T__18))) != 0):
                    self.state = 286
                    self.ebnfSuffix()


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerBlockContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lexerAltList(self):
            return self.getTypedRuleContext(jsgParserParser.LexerAltListContext,0)


        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerBlock

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerBlock" ):
                return visitor.visitLexerBlock(self)
            else:
                return visitor.visitChildren(self)




    def lexerBlock(self):

        localctx = jsgParserParser.LexerBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_lexerBlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 291
            self.match(jsgParserParser.T__13)
            self.state = 292
            self.lexerAltList()
            self.state = 293
            self.match(jsgParserParser.T__14)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LexerAtomContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def terminal(self):
            return self.getTypedRuleContext(jsgParserParser.TerminalContext,0)


        def LEXER_CHAR_SET(self):
            return self.getToken(jsgParserParser.LEXER_CHAR_SET, 0)

        def ANY(self):
            return self.getToken(jsgParserParser.ANY, 0)

        def getRuleIndex(self):
            return jsgParserParser.RULE_lexerAtom

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLexerAtom" ):
                return visitor.visitLexerAtom(self)
            else:
                return visitor.visitChildren(self)




    def lexerAtom(self):

        localctx = jsgParserParser.LexerAtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 52, self.RULE_lexerAtom)
        try:
            self.state = 298
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jsgParserParser.LEXER_ID, jsgParserParser.STRING]:
                self.enterOuterAlt(localctx, 1)
                self.state = 295
                self.terminal()
                pass
            elif token in [jsgParserParser.LEXER_CHAR_SET]:
                self.enterOuterAlt(localctx, 2)
                self.state = 296
                self.match(jsgParserParser.LEXER_CHAR_SET)
                pass
            elif token in [jsgParserParser.ANY]:
                self.enterOuterAlt(localctx, 3)
                self.state = 297
                self.match(jsgParserParser.ANY)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TerminalContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LEXER_ID(self):
            return self.getToken(jsgParserParser.LEXER_ID, 0)

        def STRING(self):
            return self.getToken(jsgParserParser.STRING, 0)

        def getRuleIndex(self):
            return jsgParserParser.RULE_terminal

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTerminal" ):
                return visitor.visitTerminal(self)
            else:
                return visitor.visitChildren(self)




    def terminal(self):

        localctx = jsgParserParser.TerminalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 54, self.RULE_terminal)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 300
            _la = self._input.LA(1)
            if not(_la==jsgParserParser.LEXER_ID or _la==jsgParserParser.STRING):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import sys

from pyjsg.parser_impl import __version__, __license__, __url__

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

requires = ["antlr4-python3-runtime>=4.7", "jsonasobj>=1.2.0", "requests"]

if sys.version_info < (3, 7):
    requires.append("dataclasses")


setup(
    name='PyJSG',
    version=__version__,
    packages=['pyjsg.jsglib', 'pyjsg.parser', 'pyjsg.parser_impl'],
    url=__url__,
    license=__license__,
    author='Harold Solbrig',
    author_email='solbrig.harold@mayo.edu',
    description='PyJSG - Python JSON Schema Grammar bindings',
    long_description='A tool to create Python classes that represent JSON objects defined in JSG',
    install_requires=requires,
    tests_require = ["yadict-compare>=1.1.2"],
    scripts=['scripts/generate_parser'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Code Generators',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.7']
)



from io import StringIO
from typing import Optional, TextIO, cast, Any


class Logger:
    """
    Basic error recording utility.  Used for validation routines where callers may simply wish to know whether something
    is or isn't valid or, in other circumstances, may wish to see the detail of all of the violations.  The mode is
    controlled by the presence of logfile.  If it is present, errors are logged and all errors are recorded.  If
    absent, the fact that the error exists is noted.
    """
    def __init__(self, logfile: Optional[TextIO] = None):
        """
        Construct a logging instance
        :param logfile: File to log to.  If absent, no messages are recorded
        """
        self.nerrors = 0
        self._logfile = logfile

    def log(self, txt: str) -> bool:
        """ Log txt (if any) to the log file (if any). Return value indicates whether it is ok to terminate on the first
        error or whether we need to continue processing.

        :param txt: text to log.
        :return: True if we aren't logging, False if we are.
        """
        self.nerrors += 1
        if self._logfile is not None:
            print(txt, file=self._logfile)
        return not self.logging

    @staticmethod
    def json_repr(item: Any) -> str:
        return f"'{item}'" if isinstance(item, str) else item

    @property
    def logging(self):
        """ Return True if errors are being recorded in the log, false if just checking for anything wrong
        """
        return self._logfile is not None

    def getvalue(self) -> Optional[str]:
        """ Return the current contents of the log file, if any """
        return self._logfile.read() if self._logfile else None


def logger():
    """ Return a """
    return Logger(cast(TextIO, StringIO()))



from typing import Optional, List, Tuple

from pyjsg.parser.jsgParser import *
from pyjsg.parser.jsgParserVisitor import jsgParserVisitor
from pyjsg.parser_impl.jsg_doc_context import JSGDocContext, PythonGeneratorElement
from pyjsg.parser_impl.parser_utils import flatten_unique


class JSGArrayExpr(jsgParserVisitor, PythonGeneratorElement):
    def __init__(self, context: JSGDocContext, ctx: Optional[jsgParser.ArrayExprContext] = None):
        from pyjsg.parser_impl.jsg_valuetype_parser import JSGValueType
        from pyjsg.parser_impl.jsg_ebnf_parser import JSGEbnf

        self._context = context
        self._types: List[JSGValueType] = None
        self._ebnf: JSGEbnf = JSGEbnf(context)
        self._ebnf.min = 0
        self._ebnf.max = None
        self.text = ""

        if ctx:
            self.text = ctx.getText()
            self.visit(ctx)

    def __str__(self):
        type_list = ' | '.join([str(t) for t in self._types])
        if len(self._types) != 1:
            type_list = f'({type_list})'
        return f"arrayExpr: [{type_list}{self._ebnf}]"

    def python_type(self) -> str:
        type_list = ', '.join([t.python_type() for t in self._types])
        if len(self._types) > 1:
            type_list = f'typing.Union[{type_list}]'
        return f"typing.List[{type_list}]"

    def _inner_signature(self) -> str:
        type_list = ', '.join([t.signature_type() for t in self._types])
        if len(self._types) > 1:
            type_list = f'typing.Union[{type_list}]'
        return type_list

    def signature_type(self) -> str:
        return f"jsg.ArrayFactory('{{name}}', _CONTEXT, {self._inner_signature()}, {self._ebnf.min}, {self._ebnf.max})"

    def reference_type(self) -> str:
        return self.signature_type()

    def mt_value(self) -> str:
        return "None"

    def members_entries(self, all_are_optional: Optional[bool] = False) -> List[Tuple[str, str]]:
        return []

    def dependency_list(self) -> List[str]:
        return flatten_unique([t.dependency_list() for t in self._types])

    # ***************
    #   Visitors
    # ***************
    def visitArrayExpr(self, ctx: jsgParser.ArrayExprContext):
        """ arrayExpr: OBRACKET valueType (BAR valueType)* ebnfSuffix? CBRACKET; """
        from pyjsg.parser_impl.jsg_ebnf_parser import JSGEbnf

        from pyjsg.parser_impl.jsg_valuetype_parser import JSGValueType
        self._types = [JSGValueType(self._context, vt) for vt in ctx.valueType()]
        if ctx.ebnfSuffix():
            self._ebnf = JSGEbnf(self._context, ctx.ebnfSuffix())



from typing import Dict, Any, Union, ForwardRef, Callable, Type
from collections.abc import Iterable


# Typing_patch module for Python 3.7


def proc_forward(etype, namespace: Dict[str, Any]):
    """ Resolve etype to an actual type if it is a forward reference """
    return etype._evaluate(namespace, namespace) if type(etype) is ForwardRef else etype


def is_union(etype) -> bool:
    """ Determine whether etype is a Union """
    return getattr(etype, '__origin__', None) is not None and \
           getattr(etype.__origin__, '_name', None) and\
           etype.__origin__._name == 'Union'


def is_dict(etype) -> bool:
    """ Determine whether etype is a Dict """
    return issubclass(type(etype), dict)


def is_iterable(etype) -> bool:
    """ Determine whether etype is a List or other iterable """
    return getattr(etype, '__origin__', None) is not None and issubclass(etype.__origin__, Iterable)


def union_conforms(element: Union, etype, namespace: Dict[str, Any], conforms: Callable) -> bool:
    """ Determine whether element conforms to at least one of the types in etype

    :param element: element to test
    :param etype: type to test against
    :param namespace: Namespace to use for resolving forward references
    :param conforms: conformance test function
    :return: True if element conforms to at least one type in etype
    """
    return any(conforms(element, t, namespace) for t in etype.__args__)



import sys

from .conformance import conforms
from .jsg_strings import JSGPattern, JSGString, Boolean, Integer, Number, String, JSGPatternedValMeta
from .jsg_anytype import AnyType, AnyTypeFactory
from .jsg_array import JSGArray,ArrayFactory
from .jsg_object import JSGObject, Object, ObjectFactory
from .jsg_objectmap import JSGObjectMap
from .loader import isinstance_, load, loads
from .jsg_validateable import JSGValidateable
from .empty import Empty
from .jsg_null import JSGNull
from .jsg_context import JSGContext

if sys.version_info < (3, 7):
    from typing import _ForwardRef as ForwardRef
    from .typing_patch_36 import *
else:
    from typing import ForwardRef
    from .typing_patch_37 import *
